:- dynamic
    apply3579/3,
    node3579/4,
    nf3579/2,
    indexCount/1.


%% getPredOrder(+, +, +, -)
%% getPredOrder(P1, P2, L, O)
%% True when P1 < P2 and O = lt, or P1 > P2 and O = gt, or P1 = P2 and O = eq. 
%% The <>= relation is defined by the user defined list L. The predicate name 
%% appearind earlier in the list is smaller. 

getPredOrder(P, P, _, eq) :-
    !.
getPredOrder(P1, _, [P1 | _], lt) :- 
    !.
getPredOrder(_, P2, [P2 | _], gt) :- 
    !.
getPredOrder(P1, P2, [_ | Rest], O) :-
    getPredOrder(P1, P2, Rest, O).



domVarOrder(V1, V2, Ord) :- 
    name(V1, [118 | A]),
    name(V2, [118 | B]),
    name(X, A),
    name(Y, B),
    (X < Y -> Ord = lt ; Ord = gt).




%% getVarOrder(+, +, -)
%% getVarOrder(V1, V2, O)
%% True when O expresses the relation (gt, lt or eq) of V1 to V2. Domain constants 
%% are smaller than action parameters that are smaller that domain variables. If 
%% both V1 and V2 are of the same type then the ordering is lexicographical. 

getVarOrder1(V, V, eq) :- 
    !.
getVarOrder1(V1, V2, Ord) :- 
    domConst(V1),
    (once(domConst(V2)),
     V1 @> V2,
     Ord = gt
     ;
     Ord = lt),
    !.
getVarOrder1(_, V2, gt) :- 
    domConst(V2),
    !.
getVarOrder1(V1, V2, Ord) :- 
    domVar(V1),
    (domVar(V2),
     domVarOrder(V1, V2, Ord)
     ;
     Ord = lt),
    !.
getVarOrder1(_, V2, gt) :- 
    domVar(V2),
    !.
getVarOrder1(V1, V2, lt) :-
    V1 @< V2,
    !.
getVarOrder1(V1, V2, gt) :- 
    V1 @> V2.



getVarOrder(V1, V2, Ord) :-
    once(getVarOrder1(V1, V2, Ord1)),
%    getVarOrder1(V1, V2, Ord1),
    !,
    Ord = Ord1.

%% getVarListOrder(+, +, -)
%% getVarListOrder(L1, L2, O)
%% True when O describes the relation (gt, lt, or eq) between lists L1 and L2 of 
%% literal parameters. The relation between individual elements is decided by 
%% getVarOrder(+, +, -).

getVarListOrder([], [], eq) :- 
    !.
getVarListOrder([V1 | Rest1], [V2 | Rest2], Order) :- 
    getVarOrder(V1, V2, Ord),
    (Ord = eq,
     !,
     getVarListOrder(Rest1, Rest2, Order)
     ;
     Order = Ord).



%% order(+, +, -)
%% order(L1, L2, O)
%% True when L1 and L2 are terms and O describes the relation between them. 

order(L1, L2, gt) :- 
    number(L1), 
    not(number(L2)),
    !.
order(L1, L2, lt) :- 
    number(L2),
    not(number(L1)),
    !.
order(L1, L2, Order) :- 
    not(number(L1)),
    not(number(L2)),
    L1 =.. [P1 | V1],
    L2 =.. [P2 | V2],
    predOrder(Ord),
    getPredOrder(P1, P2, Ord, PO),
    !,
    (PO = eq,
     !,
     getVarListOrder(V1, V2, Order)
     ;
     Order = PO).




%% applyOp(+, +, +, -)
%% applyOp(N1, N2, Op, N)
%% True when N is the result of applying the numerical operation Op on N1 and N2. 
    
applyOp(N1, N2, +, N) :-
    N is N1 + N2,
    !.
applyOp(N1, N2, -, N) :- 
    N3 is N1 - N2, 
    maxNum(N3, 0.0, N),
    !.
applyOp(N1, N2, *, N) :-
    N is N1 * N2,
    !.
applyOp(N1, N2, max, N) :-
    maxNum(N1, N2, N),
    !.
applyOp(N1, N2, min, N) :-
    min(N1, N2, N),
    !.
applyOp(1.0, 1.0, and, 1.0) :-
    !.
%applyOp(1.0, 0.0, and, 0.0) :- 
%    !.
applyOp(0.0, 1.0, and, 0.0) :- 
    !.
applyOp(0.0, 0.0, and, 0.0) :-
    !.
applyOp(_, _, and, 0.0).



%%%%%%%%%%%%%%   apply1 SHORTCUTS   %%%%%%%%%%%%%%%%%%%
% None as of now
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%% apply1(+, +, +, +, +, -)
%% apply1(N1, D1, N2, D2, Op, I)
%% True when I is the index of the root of the FODD generated as a result 
%% of running the apply operation on the subtrees of D1 and D2 rooted at 
%% N1 and N2 resp. The operator for the apply is Op

apply1(node(I1, N1, _, _), _, node(I2, N2, _, _), _, _, I) :- 
%    write('one '), write('I1 = '), write(I1), write(' I2 = '), write(I2), nl,
%   write('N1 = '), write(N1), write('  I1 = '), write(I1), write('  N2 = '), write(N2), write('  I2 = '), write(I2), nl, 
    apply3579(I1, I2, node(I, _, _, _)),
%    write(I), write(' returned'), nl, 
    !.
apply1(node(I1, N1, -1, -1), _, node(I2, N2, -1, -1), _, Op, I) :-
%    write('two '), write('I1 = '), write(I1), write(' I2 = '), write(I2), nl,
    applyOp(N1, N2, Op, D1),
    leafprecision(LP),
    (LP is 0,
     D is D1
     ; 
     ((D1 = 0.0 ; D1 = 1.0), D = D1 ; float_precision(D1, LP, D))),
    (apply3579(_, _, node(I, D, -1, -1)),
%     abs((D2 - D)) =< 0.0,
     D3 = D
     ;
     D = 0.0,
     D3 = D,
     I = 0
     ;
     D = 1.0,
     D3 = D,
     I = 1
     ;
     getIndexCount(I),
     D3 = D),
    asserta(apply3579(I1, I2, node(I, D3, -1, -1))),
%    write(D3), write(' is a the leaf'), nl, 
%    write(I), write(' returned'), nl,
    !.
apply1(node(I1, N1, L1, R1), D1, node(I2, N2, L2, R2), D2, Op, I) :- 
%    write('three '), 
%    write('I1 = '), write(I1), write(' I2 = '), write(I2), nl,
    order(N1, N2, lt),
    member(node(L1, Llit, Llt, Lrt), D1),
    member(node(R1, Rlit, Rlt, Rrt), D1),
    apply1(node(L1, Llit, Llt, Lrt), D1, node(I2, N2, L2, R2), D2, Op, L),
    apply1(node(R1, Rlit, Rlt, Rrt), D1, node(I2, N2, L2, R2), D2, Op, R),
    (L = R, 
     I = L,
     apply3579(_, _, node(I, N, IL, IR)),
     asserta(apply3579(I1, I2, node(I, N, IL, IR)))
     ;
     apply3579(_, _, node(I, N1, L, R)),
     asserta(apply3579(I1, I2, node(I, N1, L, R)))
     ;
     getIndexCount(I),
     asserta(apply3579(I1, I2, node(I, N1, L, R)))),
%    write(I), write(' returned'), nl,
    !.
apply1(node(I1, N1, L1, R1), D1, node(I2, N2, L2, R2), D2, Op, I) :- 
%   write('four '), write('I1 = '), write(I1), write(' I2 = '), write(I2), nl,
    order(N1, N2, gt),
    member(node(L2, Llit, Llt, Lrt), D2),
    member(node(R2, Rlit, Rlt, Rrt), D2),
    apply1(node(I1, N1, L1, R1), D1, node(L2, Llit, Llt, Lrt), D2, Op, L),
    apply1(node(I1, N1, L1, R1), D1, node(R2, Rlit, Rlt, Rrt), D2, Op, R), 
    (L = R, 
     I = L,
     apply3579(_, _, node(I, N, IL, IR)),
     asserta(apply3579(I1, I2, node(I, N, IL, IR)))
     ;
     apply3579(_, _, node(I, N2, L, R)),
     asserta(apply3579(I1, I2, node(I, N2, L, R)))
     ;
     getIndexCount(I),
     asserta(apply3579(I1, I2, node(I, N2, L, R)))),
%    write(I), write(' returned'), nl,
    !.
apply1(node(I1, N1, L1, R1), D1, node(I2, N2, L2, R2), D2, Op, I) :- 
%    write('five '), write('I1 = '), write(I1), write(' I2 = '), write(I2), nl, 
    order(N1, N2, eq),
    member(node(L1, Llit1, Llt1, Lrt1), D1),
    member(node(R1, Rlit1, Rlt1, Rrt1), D1),
    member(node(L2, Llit2, Llt2, Lrt2), D2),
    member(node(R2, Rlit2, Rlt2, Rrt2), D2),
    apply1(node(L1, Llit1, Llt1, Lrt1), D1, node(L2, Llit2, Llt2, Lrt2), D2, Op, L),
    apply1(node(R1, Rlit1, Rlt1, Rrt1), D1, node(R2, Rlit2, Rlt2, Rrt2), D2, Op, R), 
    (L = R, 
     I = L,
     apply3579(_, _, node(I, N, IL, IR)),
     asserta(apply3579(I1, I2, node(I, N, IL, IR)))
     ; 
     apply3579(_, _, node(I, N1, L, R)),
     asserta(apply3579(I1, I2, node(I, N1, L, R))) 
     ;
     getIndexCount(I),      
     asserta(apply3579(I1, I2, node(I, N1, L, R)))).
%    write(I), write(' returned'), nl.





%%%%%%% APPLY SHORTCUTS %%%%%%%%%%%%%
%%%%%%% * SHORTCUTS %%%%%%%%%%%%%%%%%
apply(_, _, node(0, 0.0, -1, -1), [node(0, 0.0, -1, -1)], *, [node(0, 0.0, -1, -1)]) :- 
    !.
apply(node(0, 0.0, -1, -1), [node(0, 0.0, -1, -1)], _, _, *, [node(0, 0.0, -1, -1)]) :- 
    !.
apply(_, D, node(1, 1.0, -1, -1), [node(1, 1.0, -1, -1)], *, D) :- 
    !.
apply(node(1, 1.0, -1, -1), [node(1, 1.0, -1, -1)], _, D, *, D) :- 
    !.


%%%%%% + SHORTCUTS %%%%%%%%%%%%%%%%%
apply(_, D, node(0, 0.0, -1, -1), [node(0, 0.0, -1, -1)], +, D) :- 
    !.
apply(node(0, 0.0, -1, -1), [node(0, 0.0, -1, -1)], _, D, +, D) :- 
    !.


%%%%%% max SHORTCUTS %%%%%%%%%%%%%%%%%
apply(_, D, node(0, 0.0, -1, -1), [node(0, 0.0, -1, -1)], max, D) :- 
    !.
apply(node(0, 0.0, -1, -1), [node(0, 0.0, -1, -1)], _, D, max, D) :- 
    !.



%% apply(+, +, +, +, +, -)
%% apply(N1, D1, N2, D2, Op, Res)
%% True when Res is the FODD obtained by running the apply operation on the subFODDs of 
%% D1 and D2 rooted at nodes N1 and N2 resp. and strong reducing the result. Since 
%% the resulting FODD is build bottom up using dynamic programming, indexCount(X) is 
%% used to assign indices to the generated nodes of Res.  

apply(node(I1, N1, L1, R1), D1, node(I2, N2, L2, R2), D2, Op, Res) :- 
    retractall(indexCount(_)),
    asserta(indexCount(2)),
    apply1(node(I1, N1, L1, R1), D1, node(I2, N2, L2, R2), D2, Op, _),
%    write('back from apply1'), nl, 
    retract(indexCount(_)),
    !,
    all(X, apply3579(_, _, X), L),
    retractall(apply3579(_, _, _)),
    insertsort(L, Res1),
%    write('before r5: '), write(Res1), nl, 
    !,
%    r3(Res1, Res2),
    Res1 = [node(I, _, _, _) | _],
%    (Res1 = Res2, write('yes they are equal'), nl ; write('wow r3 did something'), nl),
    !,
%    nl, write('after r3: '), write(Res2), nl, 
    (modelchecking, r1r2Cycle([I], Res1, Res) ; r5([I], Res1, Res)),
    !,
    retractall(nf3579(_, _)).
    
    
    