
:- dynamic 
    actionList/3,
    nf3579/2,
    testDomain/1,
    initState/1,
    testOps/1,
    goal/2,
    goal/1,
    rewdConj/1.

%:- use_module(library(random)).


    


get_highest_leaf_reached([], _, 0.0) :- 
    !.
get_highest_leaf_reached([path(Map, _, _, Path, _) | Rest], State, Map) :-
    not(not(subsumption(Path, State, noconv, 1))),
    !.
get_highest_leaf_reached([_ | Rest], State, Map) :- 
    get_highest_leaf_reached(Rest, State, Map).



get_highest_leaf_reached1([], _, _, 0.0) :- 
    !.
get_highest_leaf_reached1([path(Map, _, _, Path, Action) | Rest], State, Action, Map) :-
    not(not(subsumption(Path, State, noconv, 1))),
    !.
get_highest_leaf_reached1([_ | Rest], State, Action, Map) :- 
    get_highest_leaf_reached1(Rest, State, Action, Map).




get_log_of(0, 0.0) :- 
    !.
get_log_of(0.0, 0.0) :- 
    !.
get_log_of(X, Y) :- 
    Y is log(X).



get_ground_action_map1(P, Action, State, Goal, Map) :-
%    (Action = deliver(ball1,box1, loc1), trace ; true),
    findall(Leaf, (member(goal(RankValue, Lit), Goal), once((actionList(Lit, D), get_highest_leaf_reached1(D, State, Action, Leaf1), Leaf is Leaf1 * RankValue))), Maps),
%    write('Maps: '), write(Maps), nl, 
    sum_list(Maps, Map).



get_ground_action_map(P, Action, State, Goal, Map) :-
    findall(Leaf, (member(goal(RankValue, Lit), Goal), once((actionList(Action, Lit, D), get_highest_leaf_reached(D, State, Leaf1), Leaf is Leaf1 * RankValue))), Maps),
%    write('Maps: '), write(Maps), nl, 
    sum_list(Maps, Map).





get_best_ground_actions(ActSet, State, Goal, bestAct(HighestMap, ActList)) :-
    get_ground_action_maps(ActSet, State, Goal, ActList1),
    (ActList1 = [],
     HighestMap = 0.0,
     ActList = []
     ;
     (all(Leaf, member(act(Leaf, _), ActList1), Leaves) ; Leaves = []),
     maxNumList(Leaves, HighestMap),
     (all(Action, member(act(HighestMap, Action), ActList1), ActList) ; ActList = [])).
    
        
    




get_map(D, State, Map) :-
    get_leaves(D, LI1),
    insertsort(LI1, LeafIndices),
    get_highest_leaf_reached(LeafIndices, D, State, Map).




getTuple(D, 0, []).
getTuple(D, Count, [X | Rest]) :- 
    del(X, D, D1),
    Count1 is Count - 1,
    getTuple(D1, Count1, Rest).





getGroundTuple([], [], _, []) :-
    !.
getGroundTuple([X | Rest], [_ | RestTuple], Objs, [X | Rest1]) :-
    once(ground(X)),
    getGroundTuple(Rest, RestTuple, Objs, Rest1).
getGroundTuple([X | Rest], [Y | RestTuple], Objs, [X | Rest1]) :-
    once(member(objects(Y, ObjList), Objs)),
%    length(ObjList, Len),
%    random(0, Len, Rand),
%    nth0(Rand, ObjList, X), 
    member(X, ObjList),
    getGroundTuple(Rest, RestTuple, Objs, Rest1).






groundAction(Action, Action) :-
    once(ground(Action)),
    !.
groundAction(Action, GroundAction) :-
    once((Action =.. [P | Vars],
	  testOps(TestOpList),
	  member(op(P, Tuple, _), TestOpList),
	  testDomain(Objs))),
    getGroundTuple(Vars, Tuple, Objs, GroundTuple),
    once((GroundAction =.. [P | GroundTuple])).




get_best_actions([], BestActionsSoFar, BestActions) :- 
    all(Action, member(act(_, Action), BestActionsSoFar), BestActions),
    !.
get_best_actions([act(N, A) | Rest], [act(BestSoFarN, _) | RestBestSoFar], BestActions) :- 
    N > BestSoFarN,
    get_best_actions(Rest, [act(N, A)], BestActions),
    !.
get_best_actions([act(N, A) | Rest], [act(BestSoFarN, BestSoFarA) | RestBestSoFar], BestActions) :- 
    N is BestSoFarN,
    get_best_actions(Rest, [act(N, A), act(BestSoFarN, BestSoFarA) | RestBestSoFar], BestActions),
    !.
get_best_actions([_ | Rest], BestSoFar, BestActions) :- 
    get_best_actions(Rest, BestSoFar, BestActions).






getAction(State, Goal, FinalAction) :-
    
%write('--- here ---\n'),
%write('MyState('),write(State),write(').'),nl,
%write(Goal),nl,
    testOps(TestOps),
%    write('Getting Action'), nl, 
    all(act(Map, Action), (member(op(P, _, class(Action, _, preconds(PreConds), _)), TestOps), once(subsumption_all(PreConds, State, Action, ActLits)), member(Action, ActLits), groundAction(Action, Action1), once(get_ground_action_map1(P, Action1, State, Goal, Map))), AllActionSet),
    write('All Actions: '),write(AllActionSet), nl, 
    (debug(8) -> write('Optimal Actions: '), write(HighestActionSet), nl ; true),
    AllActionSet = [act(FirstN, FirstA) | RestActionSet],
    get_best_actions(RestActionSet, [act(FirstN, FirstA)], HighestActionSet),
    write('Optimal Actions: '), write(HighestActionSet), nl,    
    length(HighestActionSet, Len),
    random(0, Len, Rand),
    !,
    nth0(Rand, HighestActionSet, FinalAction).







getRewdPred(RewdPred) :- 
    reward([node(_, RewdConj, _, _), _, _]),
    functor(RewdConj, RewdPred, _).

    

getRewdConj([node(_, _, -1, -1) | _], C, C, []) :- 
    !.
getRewdConj([node(_, Lit, _, 0) | Rest], ConstsSoFar, FinalConsts, [Lit | Rest1]) :-
%    write('Rest1: '), write(Rest1), nl, 
    Lit =.. [_ | Vars],
    extractVarsFromVarConsts(Vars, Vars1),
    diff(Vars, Vars1, Consts1),
    sort(Consts1, Consts),
    batchInsertRev(Consts, ConstsSoFar, ConstsSoFar1),
%    write(getRewdConj(Rest, ConstsSoFar1, FinalConsts, Rest1)), nl, 
    getRewdConj(Rest, ConstsSoFar1, FinalConsts, Rest1).



    

getRewdConjVar(RewdConjVar, Key) :-
    reward(Rewd),
    getRewdConj(Rewd, [], Consts, RewdConj),
    getSubsKey(Consts, Key),
    variablizeFormula(RewdConj, Key, RewdConjVar).
    
    
fix_action_diagram([], []) :-
    !.
fix_action_diagram([node(0, _, -1, -1) | Rest], [node(0, 0.0, -1, -1) | Rest1]) :-
    fix_action_diagram(Rest, Rest1),
    !.
fix_action_diagram([node(1, _, -1, -1) | Rest], [node(1, 1.0, -1, -1) | Rest1]) :-
    fix_action_diagram(Rest, Rest1),
    !.
fix_action_diagram([N | Rest], [N | Rest1]) :-
    fix_action_diagram(Rest, Rest1).







get_all_leaf_node_formula([], _, []) :- 
    !.
get_all_leaf_node_formula([leaf(Map, I) | Rest], D, [leafNF(Map, LeafNF) | Other]) :- 
    collectNodeFormula(I, D, LeafNF),
    get_all_leaf_node_formula(Rest, D, Other).






storeActions1([], _, _, []) :- 
    !.
storeActions1([litD(Lit, ActionD1) | Rest], V, VKey, [litD(VarLit, Key, ActionD) | Other]) :-
    fix_action_diagram(ActionD1, ActionD2),
    get_leaves(ActionD2, L1),
    insertsort(L1, L),    
    Lit =.. [LitP | LitVars],
    getDVars(ActionD2, DVars1),
    sort(DVars1, DVars),
    batchInsertRev(LitVars, DVars, LitDVars),
    diff(LitDVars, V, Vars),
    getSubsKey(Vars, VarKey),      
    conc(VarKey, VKey, Key),
    write('Lit: '), write(Lit), nl, 
    variablizeTerm(Lit, Key, VarLit),
    variablizeDD(ActionD2, Key, ActionD3),
    write('V = '), write(V), nl, 
    !,
%    (V = [xa, xb], Lit = ont(a), trace ; true),
    get_all_leaf_node_formula(L, ActionD2, ActionD),
    retractall(nf3579(_, _)),
    write('five'), nl,   
    storeActions1(Rest, V, VKey, Other).


storeActions([]) :- 
      !.
storeActions([actionList(Action, LitD) | Rest]) :-
      Action =.. [_ | V],
      getSubsKey(V, VKey),
      variablizeTerm(Action, VKey, VarAction),
    write('Action:'), write(Action), nl, 
      storeActions1(LitD, V, VKey, ActionList),      
    write('two'), nl, 
      asserta(actionList(VarAction, ActionList)),
      storeActions(Rest).
    


 




getLiteral(_, [], _, []).
getLiteral(P, [T | Tuple], TestDom, [V | Vars]) :-
    member(objects(T, ObjList), TestDom),
    member(V, ObjList),
    getLiteral(P, Tuple, TestDom, Vars).






completeStateDescription(State, TestDom, CompleteState) :-
%    testDomain(TestDom),
    testPreds(TestPreds),
%    all(X, (member(objects(_, Objs1), TestDom), member(X, Objs1)), Objs),
    (all(eq(X,X), (member(objects(_, Objs), TestDom), member(X, Objs)), BasicEq) ; BasicEq = []),
    (all(not_eq(X, Y), (member(objects(_, Objs), TestDom), member(X, Objs), member(Y, Objs), once(X \= Y)), BasicInEq) ; BasicInEq = []),
    (all(NP, (member(pred(P, Tuple), TestPreds), getLiteral(P, Tuple, TestDom, Vars), once(NP =.. [P | Vars])), AllNeg) ; AllNeg = []),
    (all(NP, (member(P, State), once(getNeg(P, NP))), StateNeg) ; StateNeg = []),
    diff(AllNeg, StateNeg, Negs),
    flatten([State, Negs, BasicEq, BasicInEq], CompleteState).





getOutcome([Outcome], _, Outcome) :-
    !.
getOutcome([outcome(X, addList(A), delList(D)) | Rest], N, outcome(N, addList(A), delList(D))) :-
    X >= N,
    !.
getOutcome([outcome(X, addList(A), delList(D)) | Rest], N, Outcome) :-
    N1 is N - X,
    getOutcome(Rest, N1, Outcome).






getOutcome([Outcome], _, Outcome) :-
    !.
getOutcome([outcome(X, addList(A), delList(D)) | Rest], N, outcome(N, addList(A), delList(D))) :-
    X >= N,
    !.
getOutcome([outcome(X, addList(A), delList(D)) | Rest], N, Outcome) :-
    N1 is N - X,
    getOutcome(Rest, N1, Outcome).










takeAction(State, Action, Cost, NextState) :-
    Action =.. [P | Vars],
    testOps(TestOps),
    once(member(op(P, _, class(Action, Cost, preconds(PreConds), effects(Effects))), TestOps)), 
    !, 
    subsumption(PreConds, State, noconv, 1),
    member(cond(Cond, OList), Effects),
    subsumption(Cond, State, noconv, 1),
    random(N),
%    write('random: '), write(N), nl, 
    getOutcome(OList, N, outcome(_, addList(AddList), delList(DelList))),
%    diff(State, DelList, State1),
%    conc(State1, AddList, NextState).
    add_Lits_to_conv_State(AddList, State, NextState1),
    del_Lits_from_conv_State(DelList, NextState1, NextState).




rank_goals([], _, []) :- 
    !.
rank_goals([Goal | Rest], RankValue, [goal(RankValue, Goal) | Other]) :- 
     wgoparam(WGOP),
     RankValue1 is RankValue * WGOP,
%     RankValue1 is RankValue * 1.0,
    rank_goals(Rest, RankValue1, Other).



diff_goal_state([], _, []) :- 
    !.
diff_goal_state([Lit | Rest], State, Other) :- 
    dc_subsumption([Lit], State), 
    !,
    diff_goal_state(Rest, State, Other),
    !.
diff_goal_state([Lit | Rest], State, [Lit | Other]) :- 
    diff_goal_state(Rest, State, Other).






execute(_, PlanLenSoFar, -1.0, Rewd, Rewd, _) :-
    maxplanningsteps3579(MPS),
    PlanLenSoFar > MPS,
    !.
execute(State, PlanLen, PlanLen, Rewd, Rewd, Goal) :-
    subsumption(Goal, State, noconv, 1),  
    (debug(8) -> write('State: '), write(State), nl ; true),
    !.
execute(State, PlanLenSoFar, PlanLen, RewdSoFar, Rewd, Goal) :-
    diff_goal_state(Goal, State, Goal1),
%    write('Diff Goal: '), write(Goal1), nl,
    rank_goals(Goal1, 1.0, Goal2),
    write(Goal2), nl, 
    !,
    getAction(State, Goal2, Action),
    write('Action: '), write(Action), nl,
    (debug(8) -> write('Action: '), write(Action), nl, nl ; true),
    takeAction(State, Action, Cost, NextState),
    RewdSoFar1 is RewdSoFar + Cost,
    PlanLenSoFar1 is PlanLenSoFar + 1,
    !,
    execute(NextState, PlanLenSoFar1, PlanLen, RewdSoFar1, Rewd, Goal).






plan(InitState, Goal, GoalRewd, PlanLen, Rewd) :-
    testDomain(TestDom),
    completeStateDescription(InitState, TestDom, ExPath1),
    convert_Ex(ExPath1, ExPath),
    !,
    order_goals(Goal, OrderedGoal),
    !,
    execute(ExPath, 0, PlanLen, 0.0, Rewd, OrderedGoal).





rounds(Count, Success, PlanLenList, RewdList, TimeList) :-
	testRounds(TR),
    Count =< TR,
%    write('ROUND: '), write(Count), nl, 
    initState(S),
    (goal(G), GoalRewd = 0.0 ; goal(G, GoalRewd)),
    (once((statistics(cputime, [T1, _]),
	  plan(S, G, GoalRewd, PlanLen, Rewd))),
     PlanLen > 0.0,
     TotalRewd is GoalRewd + Rewd,
     statistics(cputime, [T2, _]),
     T3 is (T2 - T1)/1000,
     write('Plan Length: '), write(PlanLen), nl, nl, 
%     write('Time Taken: '), write(T3), write(' seconds'), nl, 
%     write('Reward: '), write(TotalRewd), nl, nl, nl,
     PlanLenList1 = [PlanLen | PlanLenList], 
     TimeList1 = [T3 | TimeList],
     Success1 is Success + 1,
     RewdList1 = [TotalRewd | RewdList]
     ;
     write('Plan failed'), nl, nl, nl, 
     PlanLenList1 = PlanLenList, 
     TimeList1 = TimeList,
     Success1 = Success,
     RewdList1 = [0.0 | RewdList]), 
    Count1 is Count + 1,
    !,
    rounds(Count1, Success1, PlanLenList1, RewdList1, TimeList1),
    !.
rounds(_, Success, PlanLenList, RewdList, TimeList) :-
    length(PlanLenList, Len),
    sum_list(PlanLenList, Total),
    AvgLen is Total/Len,
    length(RewdList, RLen),
    sum_list(RewdList, RTotal),
    AvgRwd is RTotal/RLen,
    length(TimeList, TLen),
    sum_list(TimeList, TTotal),
    AvgTime is TTotal/TLen,
    Percentage is Success/30,
    write('% of problems solved: '), write(Percentage), nl,
    write('Average Plan Length of Solved Problems: '), write(AvgLen), nl,
    write('Average Reward: '), write(AvgRwd), nl, 
    write('Average Time Taken: '), write(AvgTime), nl, 
    write('***************************************'), nl, nl, nl, nl.




segregateActionLists([], _, []) :- 
      !.
segregateActionLists([Action | Rest], AllActionLitLists, [actionList(Action, ActionList) | Other]) :-
      all(litD(Lit, D), member(actionList(Action, Lit, D), AllActionLitLists), ActionList),
      segregateActionLists(Rest, AllActionLitLists, Other).

    


adjustActionCostsForProblem([], TestOps, TestOps) :- 
    !.
adjustActionCostsForProblem([action(Act, Cost) | Rest], TestOps, NewTestOps) :-
    del(op(Act, X, class(Y, OldCost, Z, W)), TestOps, TestOps1),
    adjustActionCostsForProblem(Rest, [op(Act, X, class(Y, Cost, Z, W)) | TestOps1], NewTestOps).
    
    


solve_all([]) :-
    !.
solve_all([Problem | Rest]) :-
    testDomain(Problem, TD),
    initState(Problem, S),
    asserta(testDomain(TD)),
    asserta(initState(S)),
%    storeActions(ActionDiagramsList),
%    write('Actions Stored'), nl, 
    write(Problem), nl, 
    (once(cost(Problem, CostList)),
     CostList \= [],
     retract(testOps(TestOps)),
     adjustActionCostsForProblem(CostList, TestOps, NewTestOps),
     asserta(testOps(NewTestOps))
     ;
     true),   
    (goal(Problem, G),
     !,
     asserta(goal(G)),
     rounds(1, 0, [], [], []),
     retract(goal(G))
     ;
     goal(Problem, G, GoalRewd),
     !,
     asserta(goal(G, GoalRewd)),
     rounds(1, 0, [], [], []),
     retract(goal(G, RewdGoal))),
%    retract(rewdConj(_)),
%    retractall(actionList(_, _)),
    retract(testDomain(TD)),
    retract(initState(S)),
    solve_all(Rest).




solve :-
    problems(P),
%%    all(Act, actionList(Act, _, _), AllActions),
%%    (all(actionList(Action, Lit, D), retract(actionList(Action, Lit, D)), ActionDiagramsList1) ; ActionDiagramsList1 = []),    
%%    segregateActionLists(AllActions, ActionDiagramsList1, ActionDiagramsList),
%%    storeActions(ActionDiagramsList),
%    write(ActionDiagramList), nl.
    solve_all(P).

