:- dynamic
    ns3579/1,
    interpretations3579/2,
    currstates3579/1,
    firststate3579/1,
    alldomactions3579/1,
    edges/1,
    values/1,
    removeNode3579/1.
    


:- use_module(library(timeout)).


okToParallelizeNode(LenPath) :-
		(LenPath > 90 -> parallelize_node_removal). 

okToParallelizeEdge(LenPath) :-
		(LenPath > 90 -> parallelize_edge_removal). 





getHighestPathTraversedByI(_, [], [], 0.0) :- 
    !.
getHighestPathTraversedByI(I, [path(N, _, Edges, Path) | Paths], Edges, N) :-
    not(not(subsumption(Path, I, noconv, 1))),
%    not(not(dc_subsumption(Path, I))),
%    time_out(not(not(dc_subsumption(Path, I))), 1000, TO),
%    !,
%    TO = success,
%    write('I traversed: '), write(Path), nl,nl,
    !.
getHighestPathTraversedByI(I, [_ | Paths], Edges, N) :-
    getHighestPathTraversedByI(I, Paths, Edges, N).





perform_r13(_, [], Edges, Edges, []) :-
    !.
perform_r13(Paths, [I | Ilist], EdgesSoFar, Edges, [N | RestN]) :-
    (debug(r12) -> nl, write(I), nl ; true),
    (getHighestPathTraversedByI(I, Paths, PathEdges, N) ; PathEdges = []),
    batchInsert(PathEdges, EdgesSoFar, EdgesSoFar1),
    perform_r13(Paths, Ilist, EdgesSoFar1, Edges, RestN).
    


output_list([], _) :- 
	!.
output_list([E | Elements], Delim) :- 
	write(E), write(Delim),
	output_list(Elements, Delim).	



output_paths([]) :- 
    !. 	
output_paths([path(N,L,E,P) | Paths]) :-
    (all(Var, (member(Lit,P), once((Lit =.. [_ | LitVars])), member(Var, LitVars), once((domVar(Var)))), PathVars) ; PathVars = []),
    write(N), write(':'),  write(L), write(':'), 
    write(E), write(':'),
    write(P), write(':'),
    write(PathVars),
    nl,	
    output_paths(Paths).
    	
	




perform_parallel_r13(Paths, IList, Edges, Values) :- 
    tell('reduction_paths'),
    output_list(Paths, '.\n'),
	told,
	tell('reduction_interpretations'),
    output_list(IList, '.\n'),
    told, !,
		system("python parallelize_reduction.py edges 0 0"),
    %system("perl parallelize_reduction.pl edges 0 0"),
    reconsult('reductionResult.yap'), 
    !,
    retract(values(Values)), !,
    retract(edges(Edges)).







r13(D, R) :- 
    D = [node(Root, _, _, _) | _],
    get_all_edges(D, AllEdges),
    getDVarsConsts(D, [], VarsConsts),
    domainConsts(DC),
    diff(VarsConsts, DC, Vars),
    getSubsKey(Vars, Key),
    variablizeDD(D, Key, DVar),
    get_paths(Root, DVar, [], [], Paths1),
    length(Paths1, LenP1),
    insertsort(Paths1, Paths),
    length(Paths, LenP2),
%    write(Paths), nl, 
    getRewdPred(RewdPred),
    interpretations3579(RewdPred, Ilist),
		!,
		(okToParallelizeEdge(LenP2),
		!,
		perform_parallel_r13(Paths, Ilist, Edges, Values)
		;
		perform_r13(Paths, Ilist, [], Edges, Values)),
    diff(AllEdges, Edges, RemovableEdges),
%%    write('Removable Edges: '), write(RemovableEdges), nl,
    (RemovableEdges = [],
     R = D
     ;
     insert(node(0, 0.0, -1, -1), D, D0),
     replace_edge_target(D0, RemovableEdges, D1, [], HL1),
     insertsort(HL1, HL2),
    removeUnreachable(_, HL2, D1, D2, HL3), 
     once(eraseBatchNF(RemovableEdges, D)),
     once(eraseBatchSuperss(RemovableEdges, D)),
     (debug(1) -> write('calling drop cycle'), nl ; true),
     (debug(main1), write('before node removal'), nl, pp(D2), nl ; true), 
     (modelcheckingnode, r13_node_remove(D2, RemovableEdges, Ilist, Values, Key, D3) ; D3 = D2),
    R = D3).






run_Int_on_node_paths([], _, _) :- 
    !.
run_Int_on_node_paths([path(N, _, _, _) | _], _, V) :- 
    N =< V,
    !.
run_Int_on_node_paths([path(_, _, _, Path) | Paths], I, V) :- 
    !,
    once(not(subsumption(Path, I, noconv, 1))),
    run_Int_on_node_paths(Paths, I, V).
    
    
    
    



run_IList_on_node_paths(_, [], []) :- 
    !.
run_IList_on_node_paths(Paths, [I | IList], [V | Values]) :-
    !,
    run_Int_on_node_paths(Paths, I, V),
    run_IList_on_node_paths(Paths, IList, Values).
    
    




run_parallel_IList_on_node_paths(Paths, IList, Values) :- 
    tell('reduction_paths'),
    output_list(Paths, '.\n'),
    told,
    tell('reduction_interpretations'),
    output_list(IList,'.\n'),
    told , 
    tell('reduction_values'),
    output_list(Values, '.\n'),
    told,	
    !,
		system("python parallelize_reduction.py edges 0 0"),
		%system("perl parallelize_reduction.pl nodes 0 0"),
    reconsult('reductionResult.yap'), 
    retract(removeNode3579(RemoveNode)),	
    !,
    RemoveNode = 1.
	    	
    	
	
	
	
	
	
    
r13_node_remove(D, [], _, _, _, D) :- 
    retractall(nf3579(_,_)),
    !.
r13_node_remove(D, [I-Dir | Rest], IList, Values, Key, D3) :- 
    once(member(node(I, _, L, 0), D)),
    once(collectNodeFormula(I, D, NF)),    
    once(get_paths(L, D, [], [], Paths)),
    once(all(path(N, Plen, EL, Path), (get_a_disjunct(NF, [], SuperPath), member(path(N, Plen, EL, SubPath), Paths), conc(SuperPath, SubPath, Path1), variablizeFormula(Path1, Key, Path)), AllPaths1)),
    once(insertsort(AllPaths1, AllPaths)),
		length(AllPaths,LenPath),
		(okToParallelizeNode(LenPath),
		!,
		run_parallel_IList_on_node_paths(AllPaths, IList, Values)
		;
		run_IList_on_node_paths(AllPaths, IList, Values)),
    del(node(I, _, L, 0), D, D1),    
    parentUpdate(D1, L, I, D2, _), 
    retractall(nf3579(_,_)),
    once(eraseBatchNF([I-Dir], D)),
    once(eraseBatchSuperss([I-Dir], D)),    
    r13_node_remove(D2, Rest, IList, Values, Key, D3),
    !.
		
r13_node_remove(D, [I-Dir | Rest], IList, Values, Key, D3) :-
    once(member(node(I, _, 0, R), D)),
    once(collectNodeFormula(I, D, NF)),    
    once(get_paths(R, D, [], [], Paths)),
    once(all(path(N, Plen, EL, Path), (get_a_disjunct(NF, [], SuperPath), member(path(N, Plen, EL, SubPath), Paths), conc(SuperPath, SubPath, Path1), variablizeFormula(Path1, Key, Path)), AllPaths1)),
    once(insertsort(AllPaths1, AllPaths)),
		length(AllPaths,LenPath),
		(okToParallelizeNode(LenPath),
		!, 
		run_parallel_IList_on_node_paths(AllPaths, IList, Values)
		;
		run_IList_on_node_paths(AllPaths, IList, Values)),
    del(node(I, _, 0, R), D, D1),
    parentUpdate(D1, R, I, D2, _),
    retractall(nf3579(_,_)),
    once(eraseBatchNF([I-Dir], D)),
    once(eraseBatchSuperss([I-Dir], D)),
    r13_node_remove(D2, Rest, IList, Values, Key, D3),
    !.
r13_node_remove(D, [_ | Rest], IList, Values, Key, R) :- 
    r13_node_remove(D, Rest, IList, Values, Key, R).
    









